package handwriting

import (
	"bytes"
	context "context"
	fmt "fmt"
	"log"
	"net"
	"os"

	neuralnetwork "../../neuralnetwork/neuralnetwork"
	training "../../neuralnetwork/training"
	"gonum.org/v1/gonum/mat"
	grpc "google.golang.org/grpc"
)

type HandwritingService struct {
	nn *neuralnetwork.NeuralNetwork
}

func NewHandwritingService() (hws *HandwritingService) {
	hws = &HandwritingService{}
	hws.nn, _ = neuralnetwork.NewNeuralNetwork([]int{784, 300, 200, 10}, neuralnetwork.NewRPropInitializer(neuralnetwork.RPropConfig{
		NuPlus:   1.2,
		NuMinus:  0.5,
		DeltaMax: 50.0,
		DeltaMin: 0.000001,
	}))
	return
}

func (hws *HandwritingService) Recognize(ctx context.Context, matrix *Matrix) (*Result, error) {
	fmt.Printf("Recognize %v size: %v\n", len(matrix.Data), hws.nn.Sizes[0])
	dense := mat.NewDense(hws.nn.Sizes[0], 1, matrix.Data)
	index, _ := hws.nn.Predict(dense)
	fmt.Printf("Recognition result %v\n", index)
	return &Result{ResultCharacter: uint32(index)}, nil
}

func (hws *HandwritingService) SetNeuralNetworkData(ctx context.Context, nnRaw *NeuralNetworkRaw) (*None, error) {
	fmt.Println("SetNeuralNetworkData")
	r := bytes.NewReader(nnRaw.Data)
	hws.nn.LoadState(r)
	return &None{}, nil
}

func (hws *HandwritingService) GetNeuralNetworkData(context.Context, *None) (*NeuralNetworkRaw, error) {
	nnRaw := &NeuralNetworkRaw{}
	fmt.Println("SetNeuralNetworkData")
	r := bytes.NewReader(nnRaw.Data)
	hws.nn.LoadState(r)
	return nnRaw, nil
}

func (hws *HandwritingService) ReTrain(context.Context, *None) (*None, error) {
	fmt.Println("ReTrain")

	trainer := training.NewMNISTReader("./mnist.data", "./mnist.labels")
	failCount := 0
	total := 0
	trainer.Reset()
	for trainer.NextValidator() {
		total++
		dataSet, expect := trainer.GetValidator()
		index, _ := hws.nn.Predict(dataSet)
		if expect.At(index, 0) != 1.0 {
			failCount++
			// fmt.Printf("Fail: %v, %v\n\n", trainer.ValidationIndex(), expect.At(index, 0))
		}
	}
	fmt.Printf("Fail count: %v/%v\n\n", failCount, total)

	hws.nn.Train(trainer, 100)

	fmt.Println("ReTrain finished")
	outFile, err := os.OpenFile("./mnistnet.nnd", os.O_CREATE|os.O_TRUNC|os.O_WRONLY, 0666)
	if err != nil {
		log.Fatal(err)
	}
	defer outFile.Close()
	hws.nn.SaveState(outFile)
	return &None{}, nil
}

func (hws *HandwritingService) Run() {
	grpcServer := grpc.NewServer()
	RegisterHandwritingServer(grpcServer, hws)
	lis, err := net.Listen("tcp", "localhost:65001")
	if err != nil {
		fmt.Printf("Failed to listen: %v\n", err)
	}

	fmt.Printf("Listen localhost:65001\n")
	if err := grpcServer.Serve(lis); err != nil {
		fmt.Printf("Failed to serve: %v\n", err)
	}
}

func drawImage(dense *mat.Dense) {
	for i := 0; i < 28; i++ {
		for j := 0; j < 28; j++ {
			val := 0
			if dense.At(i*28+j, 0) > 0 {
				val = 1
			}
			fmt.Printf("%v ", val)
		}
		fmt.Println()
	}
}
